"""
Custome logger for PyPPL
@variables:
    LOG_FORMAT (str): The format of loggers
    LOGTIME_FORMAT (str): The format of time for loggers
    GROUP_VALUES (dict): The values for each level group
    LEVEL_GROUPS (dict): The groups of levels
    THEMES (dict): The builtin themes
    SUBLEVELS (dict): the sub levels used to limit loggers of the same type
"""
import re
import logging
from copy import copy
from functools import partial

import colorama
# Fore/Back: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
# Style: DIM, NORMAL, BRIGHT, RESET_ALL
from .config import config as default_config
from .plugin import pluginmgr

colorama.init(autoreset=False)

LOG_FORMAT = "[%(asctime)s%(message)s"
# [05-10 09:10:10    MAIN.  PYPPL] message
LOGTIME_FORMAT = "%m-%d %H:%M:%S"

GROUP_VALUES = dict(TITLE=80,
                    SUBTITLE=70,
                    STATUS=60,
                    CRITICAL=50,
                    ERROR=40,
                    WARNING=30,
                    INFO=20,
                    DEBUG=10,
                    NOTSET=0)

LEVEL_GROUPS = dict(
    TITLE=['PROCESS', 'DONE'],
    SUBTITLE=['DEPENDS'],
    STATUS=['WORKDIR', 'CACHED', 'P_DONE'],
    CRITICAL=['INFO', 'BLDING', 'SBMTING', 'RUNNING', 'JOBDONE', 'KILLING'],
    ERROR=['ERROR'],
    WARNING=['WARNING', 'RTRYING'],
    INFO=['PYPPL', 'PLUGIN', 'TIPS', 'CONFIG'],
    DEBUG=['DEBUG'],
)

THEMES = dict(green_on_black=dict(TITLE='{s.BRIGHT}{f.CYAN}',
                                  SUBTITLE='{f.MAGENTA}',
                                  STATUS='{f.YELLOW}',
                                  CRITICAL='{f.GREEN}',
                                  ERROR='{f.RED}',
                                  WARNING='{s.BRIGHT}{f.YELLOW}',
                                  DEBUG='{s.DIM}{f.WHITE}',
                                  NOTSET=''),
              blue_on_black=dict(TITLE='{s.BRIGHT}{f.CYAN}',
                                 SUBTITLE='{f.GREEN}',
                                 STATUS='{f.YELLOW}',
                                 CRITICAL='{f.BLUE}',
                                 ERROR='{f.RED}',
                                 WARNING='{s.BRIGHT}{f.YELLOW}',
                                 DEBUG='{s.DIM}{f.WHITE}',
                                 NOTSET=''),
              magenta_on_black=dict(TITLE='{s.BRIGHT}{f.GREEN}',
                                    SUBTITLE='{f.BLUE}',
                                    STATUS='{f.YELLOW}',
                                    CRITICAL='{f.MAGENTA}',
                                    ERROR='{f.RED}',
                                    WARNING='{s.BRIGHT}{f.YELLOW}',
                                    DEBUG='{s.DIM}{f.WHITE}',
                                    NOTSET=''),
              green_on_white=dict(TITLE='{s.BRIGHT}{f.BLUE}',
                                  SUBTITLE='{f.MAGENTA}',
                                  STATUS='{f.YELLOW}',
                                  CRITICAL='{f.GREEN}',
                                  ERROR='{f.RED}',
                                  WARNING='{s.BRIGHT}{f.YELLOW}',
                                  DEBUG='{s.DIM}{f.BLACK}',
                                  NOTSET=''),
              blue_on_white=dict(TITLE='{s.BRIGHT}{f.GREEN}',
                                 SUBTITLE='{f.MAGENTA}',
                                 STATUS='{f.YELLOW}',
                                 CRITICAL='{f.BLUE}',
                                 ERROR='{f.RED}',
                                 WARNING='{s.BRIGHT}{f.YELLOW}',
                                 DEBUG='{s.DIM}{f.BLACK}',
                                 NOTSET=''),
              magenta_on_white=dict(TITLE='{s.BRIGHT}{f.BLUE}',
                                    SUBTITLE='{f.GREEN}',
                                    STATUS='{f.YELLOW}',
                                    CRITICAL='{f.MAGENTA}',
                                    ERROR='{f.RED}',
                                    WARNING='{s.BRIGHT}{f.YELLOW}',
                                    DEBUG='{s.DIM}{f.BLACK}',
                                    NOTSET=''))

SUBLEVELS = {
    'CACHE_FAILED': -1,
    'CACHE_INPUT_MODIFIED': -1,
    'CACHE_OUTPUT_MODIFIED': -1,
    'INFILE_RENAMING': -1,
    'INFILE_EMPTY': -1,
    'SUBMISSION_FAIL': -3,
    'OUTFILE_NOT_EXISTS': -1,
    'OUTDIR_CREATED_AFTER_RESET': -1,
    'SCRIPT_UPDATED': -2,
    'JOB_RESETTING': -1,
}


def get_group(level):
    """@API
    Get the group name of the level
    @params:
        level (str): The level, should be UPPERCASE
    @returns:
        (str): The group name
    """
    for group, levels in LEVEL_GROUPS.items():
        if level in levels:
            return group
    return 'NOTSET'


def get_value(level):
    """@API
    Get the value of the level
    @params:
        level (str): The level, should be UPPERCASE
    @returns:
        (int): The value of the group where the level is in.
    """
    if level[:1] == '_':
        return max(GROUP_VALUES.values())
    return GROUP_VALUES.get(get_group(level), 0)


class Theme:  # pylint: disable=too-few-public-methods
    """@API
    The theme for the logger
    @variables:
        COLORS (dict): Color collections used to format theme
    """
    COLORS = dict(
        Style=colorama.Style,
        s=colorama.Style,
        Back=colorama.Back,
        b=colorama.Back,
        Fore=colorama.Fore,
        f=colorama.Fore,
    )

    def __init__(self, theme='green_on_black'):
        """@API
        Construct for Theme
        @params:
            theme (str): the name of the theme
        """
        if theme is True:
            theme = 'green_on_black'
        if not theme:
            self.theme = {}
        elif isinstance(theme, dict):
            self.theme = theme
        elif theme in THEMES:
            self.theme = THEMES[theme]
        else:
            raise ValueError('No such theme: %s' % theme)

    def get_color(self, level):
        """@API
        Get the color for a given level
        @params:
            `level`: The level
        @returns:
            The color of the level by the theme.
        """
        group = get_group(level.upper())
        return self.theme.get(group, '').format(**Theme.COLORS)


class StreamFormatter(logging.Formatter):
    """
    Logging formatter for stream (sys.stderr)
    """
    def __init__(self, theme):
        logging.Formatter.__init__(self, LOG_FORMAT, LOGTIME_FORMAT)
        self.theme = theme

    def format(self, record):
        if hasattr(record, 'formatted') and record.formatted:
            return record.formatted

        # save the formatted, for all handlers
        record.msg = str(record.msg)
        if '\n' in record.msg:
            record.tails = []
            msgs = record.msg.splitlines()
            record.msg = msgs[0]
            for msg in msgs[1:]:
                rec = copy(record)
                rec.msg = msg
                self.format(rec)
                record.tails.append(rec)

        proc = record.proc
        proc += (': '
                 if hasattr(record, 'proc') and record.proc
                 else '')
        jobs = ('' if record.jobidx is None
                else '[{ji}/{jt}] '.format(
                    ji=str(record.jobidx + 1).zfill(len(str(record.joblen))),
                    jt=record.joblen
                ))
        color = self.theme.get_color(record.mylevel)
        record.msg = (f" {color}{record.plugin:>7s}."
                      f"{record.mylevel:>7s}{colorama.Style.RESET_ALL}] "
                      f"{color}{proc}{jobs}{record.msg}"
                      f"{colorama.Style.RESET_ALL}")

        setattr(record, 'formatted', logging.Formatter.format(self, record))
        return record.formatted


class StreamHandler(logging.StreamHandler):
    """
    Logging handler for stream (sys.stderr)
    """
    DATA = {'prevbar': None, 'done': {}}

    def _emit(self, record, terminator="\n"):
        prevterm = self.terminator
        self.terminator = terminator
        super(StreamHandler, self).emit(record)
        self.terminator = prevterm

    def emit(self, record):
        if record.ispbar:
            if record.done:
                proc = record.proc if hasattr(record, 'proc') else ''
                with self.lock:
                    # make sure done pbar is only shown once.
                    if not StreamHandler.DATA['done'].get(proc):
                        self._emit(record, '\n')
                        # clear previous pbars if any
                        StreamHandler.DATA['prevbar'] = None
                        StreamHandler.DATA['done'][proc] = True
            else:
                self._emit(record, '\r')
                StreamHandler.DATA['prevbar'] = record
        else:
            pbarlog = StreamHandler.DATA['prevbar']
            if pbarlog:
                formatted = re.sub(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])',
                                   '', pbarlog.formatted)
                self.stream.write(' ' * len(formatted) + '\r')

            self._emit(record, '\n')
            if hasattr(record, 'tails'):
                for tail in record.tails:
                    self._emit(tail, '\n')

            if pbarlog:
                self._emit(pbarlog, '\r')


class StreamFilter(logging.Filter):  # pylint: disable=too-few-public-methods
    """
    Logging filter for stream (sys.stderr)
    """
    def __init__(self, name, levels):
        super(StreamFilter, self).__init__(name)
        self.levels = levels
        self.subs = {}

    def filter(self, record): # pylint: disable=too-many-return-statements
        # logging is disabled
        if not self.levels:
            return False

        level = record.mylevel
        slevel = record.slevel if hasattr(record, 'slevel') else None
        proc = record.proc if hasattr(record, 'proc') else ''

        # user logs
        if level[0] == '_':
            return True
        if level not in self.levels:
            return False
        if not slevel or slevel not in SUBLEVELS:
            return True
        if not proc:
            return False

        # the limitation is only for one process
        if proc not in self.subs:
            self.subs = {
                proc: dict(zip(SUBLEVELS.keys(), [0] * len(SUBLEVELS)))
            }

        self.subs[proc][slevel] += 1
        allowed_lines = abs(SUBLEVELS[slevel])
        print_summary = SUBLEVELS[slevel] < 0
        if self.subs[proc][slevel] > allowed_lines:
            return False
        if self.subs[proc][slevel] < allowed_lines:
            return True
        # ==
        if print_summary:
            record.msg += "\n...... max={max} ({slevel}) reached".format(
                max=allowed_lines, slevel=slevel)
            record.msg += ", further information will be ignored."
        return True


class FileFilter(StreamFilter):  # pylint: disable=too-few-public-methods
    """
    Logging filter for file
    """
    def filter(self, record):
        if (record.ispbar
                and not record.done) or not hasattr(record, 'formatted'):
            return False
        return super(FileFilter, self).filter(record)


class FileFormatter(logging.Formatter):
    """
    Logging formatter for file,
    Extends StreamFormatter, removes the terminal colors
    """
    def __init__(self):
        logging.Formatter.__init__(self, LOG_FORMAT, LOGTIME_FORMAT)
        self.ansi_regex = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def format(self, record):
        # record has already been formatted by StreamFormatter
        # just remove the colors
        if hasattr(record, 'formatted'):
            return self.ansi_regex.sub('', record.formatted)
        return super(FileFormatter, self).format(record)


def init_levels(group, leveldiffs):
    """@API
    Initiate the levels, get real levels.
    @params:
        group (str): The group of levels
        leveldiffs (str|list): The diffs of levels
    @returns:
        (set): The real levels.
    """
    value = GROUP_VALUES.get(group, GROUP_VALUES['INFO'])
    groups = [key for key in GROUP_VALUES if GROUP_VALUES[key] >= value]
    ret = set(level for group in groups
              for level in LEVEL_GROUPS.get(group, []))

    if not leveldiffs:
        return ret
    if not isinstance(leveldiffs, (tuple, list, set)):
        leveldiffs = set([leveldiffs])
    for level in leveldiffs:
        prefix = level[:1] if level[:1] in ('+', '-') else '+'
        level = level[1:] if level[:1] in ('+', '-') else level
        level = level.upper()
        if prefix == '-' and level in ret:
            ret.remove(level)
        elif prefix == '+':
            ret.add(level)
    return ret


class Logger:
    """@API
    A wrapper of logger

    All plugins or main process should initialize a logger like:
    ```
    logger = Logger(plugin="Main")

    ```
    """

    __slots__ = ('_plugin', '_name', '_ispbar', '_logger')

    def __init__(self, name: str = 'PyPPL', plugin: str = 'main'):
        """@API
        The logger wrapper construct
        @params:
            name (str): The logger name. Default: `PyPPL`
            bake (dict): The arguments to bake a new logger.
        """
        self._name = name
        self._plugin = plugin.upper()
        self._ispbar = False
        self._logger = logging.getLogger(self._name)
        if self._plugin == "MAIN":
            self.initialize()

    def initialize(self, config=None):
        """@API
        Initiate the logger, called by the construct,
        Just in case, we want to change the config and as default_config
        Reinitiate the logger.
        @params:
            conf (Config): The configuration used to initiate logger.
        """

        config = config or default_config.logger
        config2 = default_config.logger.copy()
        config2.update(config)
        config = config2

        # Set a low level level, let me control the logging level acutally
        self._logger.setLevel(1)
        # Remove existing logger handlers
        for handler in self._logger.handlers:
            handler.close()
        del self._logger.handlers[:]

        # Plugins to add levels
        pluginmgr.hook.logger_init(logger=self)

        theme = Theme(config.theme)
        levels = init_levels(config.level.upper(), config.leveldiffs)
        stream_handler = StreamHandler()
        stream_handler.addFilter(StreamFilter(self._name, levels))
        stream_handler.setFormatter(StreamFormatter(theme))
        self._logger.addHandler(stream_handler)

        if config.file:
            file_handler = logging.FileHandler(config.file)
            file_handler.addFilter(FileFilter(self._name, levels))
            file_handler.setFormatter(FileFormatter())
            self._logger.addHandler(file_handler)

    def add_level(self, level, group='INFO'):  # pylint: disable=no-self-use
        """@API
        @params:
            level (str): The log level name
                Make sure it's less than 7 characters
            group (str): The group the level is to be added
        """
        level = level.upper()
        group = group.upper()
        if group not in LEVEL_GROUPS:
            raise ValueError(
                'No such level group: {}, available ones are: {}'.format(
                    group, list(LEVEL_GROUPS.keys())))
        if level not in LEVEL_GROUPS[group]:
            LEVEL_GROUPS[group].append(level)

    def add_sublevel(self, slevel, lines=-1):  # pylint: disable=no-self-use
        """@API
        @params:
            slevel (str): The debug level
            lines (int): The number of lines allowed for the debug level
                - Negative value means a summary will be printed
        """
        SUBLEVELS[slevel.upper()] = lines

    @property
    def pbar(self):
        """@API
        Mark the record as a progress record.
        Allow `logger.pbar.info` access
        @returns:
            (Logger): The Logger object itself
        """
        self._ispbar = True
        return self

    def _emit(self, *args, **kwargs):
        extra = {'jobidx': None, 'proc': '', 'done': False}
        extra.update({'mylevel': kwargs.pop('_level')})
        extra.update(kwargs.pop('_extra'))
        extra.update(kwargs.pop('extra', {}))
        extra.update(kwargs)
        # plugin is not supposed to be an argument
        # overwriting all that passed in
        extra['plugin'] = self._plugin
        self._logger.info(*args, extra=extra)

    def __getattr__(self, name):
        """@API
        Allows logger.info way to specify the level
        @params:
            name (str): The level name.
        @returns:
            (callable): The logger with the level
        """
        ispbar = self._ispbar
        self._ispbar = False
        return partial(self._emit,
                       _level=name.upper(),
                       _extra=dict(ispbar=ispbar, slevel=None))

    __getitem__ = __getattr__


# pylint: disable=invalid-name
logger = Logger()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['console', 'pyppl', 'pyppl.console']

package_data = \
{'': ['*']}

install_requires = \
['attr_property',
 'attrs>=19.3,<20.0',
 'cmdy',
 'colorama>=0.4.1,<0.5.0',
 'diot',
 'filelock>=3.0.0,<4.0.0',
 'liquidpy',
 'pluggy<1.0.0',
 'psutil>=5.6,<6.0',
 'pyparam',
 'pyppl_echo',
 'pyppl_export',
 'pyppl_lock',
 'pyppl_rich',
 'pyppl_runners',
 'pyppl_strict',
 'python-simpleconf',
 'python-varname',
 'toml>=0.10,<0.11',
 'transitions>=0.7,<0.8']

entry_points = \
{'console_scripts': ['pyppl = pyppl.console:main']}

setup_kwargs = {
    'name': 'pyppl',
    'version': '3.2.2',
    'description': 'A Python PiPeLine framework',
    'long_description': '# [PyPPL][3] - A [Py](#)thon [P](#)i[P](#)e[L](#)ine framework\n\n[![Pypi][10]][18] [![Github][11]][3] [![PythonVers][14]][18] [![docs][19]][1] [![Travis building][5]][7] [![Codacy][4]][8] [![Codacy coverage][6]][8]\n\n[Documentation][1] | [API][2] | [Change log][9]\n\n<!-- toc -->\n## Features\n- Process caching.\n- Process reusability.\n- Process error handling.\n- Runner customization.\n- Easy running profile switching.\n- Plugin system.\n\n## Installation\n```bash\npip install PyPPL\n```\n\n## Plugin gallery\n\n(*) shipped with `PyPPL`\n\n- [pyppl_report](https://github.com/pwwang/pyppl_report): Generating reports for PyPPL pipelines\n- [pyppl_flowchart](https://github.com/pwwang/pyppl_flowchart): Generating flowchart for PyPPL\n- [pyppl_export](https://github.com/pwwang/pyppl_export)*: Exporting outputs generated by PyPPL pipeline\n- [pyppl_echo](https://github.com/pwwang/pyppl_echo)*: Echoing script output to PyPPL logs"\n- [pyppl_rich](https://github.com/pwwang/pyppl_rich)*: Richer information in logs for PyPPL\n- [pyppl_strict](https://github.com/pwwang/pyppl_strict)*: More strict check of job success for PyPPL\n- [pyppl_lock](https://github.com/pwwang/pyppl_lock)*: Preventing running processes from running again for PyPPL\n- [pyppl_annotate](https://github.com/pwwang/pyppl_annotate): Adding long description/annotation for processes\n- [pyppl_require](https://github.com/pwwang/pyppl_require): Checking and installing requirements for processes\n- [pyppl_jobtime](https://github.com/pwwang/pyppl_jobtime): Job running time statistics for PyPPL\n- [pyppl_notify](https://github.com/pwwang/pyppl_notify): Email notifications for PyPPL\n- [pyppl_runcmd](https://github.com/pwwang/pyppl_runcmd): Allowing to run local command before and after each process for PyPPL\n- [pyppl_runners](https://github.com/pwwang/pyppl_runners): Common runners for PyPPL\n\n## Writing pipelines with predefined processes\nLet\'s say we are implementing the [TCGA DNA-Seq Re-alignment Workflow][16]\n(The very left part of following figure).\nFor demonstration, we will skip the QC and the co-clean parts here.\n\n[![DNA_Seq_Variant_Calling_Pipeline][15]][16]\n\n`demo.py`:\n```python\nfrom pyppl import PyPPL, Channel\n# import predefined processes\nfrom TCGAprocs import pBamToFastq, pAlignment, pBamSort, pBamMerge, pMarkDups\n\n# Load the bam files\npBamToFastq.input = Channel.fromPattern(\'/path/to/*.bam\')\n# Align the reads to reference genome\npAlignment.depends = pBamToFastq\n# Sort bam files\npBamSort.depends = pAlignment\n# Merge bam files\npBamMerge.depends = pBamSort\n# Mark duplicates\npMarkDups.depends = pBamMerge\n# Export the results\npMarkDups.config.export_dir = \'/path/to/realigned_Bams\'\n# Specify the start process and run the pipeline\nPyPPL().start(pBamToFastq).run()\n```\n\n[![asciicast][20]][21]\n\n## Implementing individual processes\n`TCGAprocs.py`:\n```python\nfrom pyppl import Proc\npBamToFastq = Proc(desc = \'Convert bam files to fastq files.\')\npBamToFastq.input = \'infile:file\'\npBamToFastq.output = [\n    \'fq1:file:{{i.infile | stem}}_1.fq.gz\',\n    \'fq2:file:{{i.infile | stem}}_2.fq.gz\']\npBamToFastq.script = \'\'\'\nbamtofastq collate=1 exclude=QCFAIL,SECONDARY,SUPPLEMENTARY \\\n    filename= {{i.infile}} gz=1 inputformat=bam level=5 \\\n    outputdir= {{job.outdir}} outputperreadgroup=1 tryoq=1 \\\n    outputperreadgroupsuffixF=_1.fq.gz \\\n    outputperreadgroupsuffixF2=_2.fq.gz \\\n    outputperreadgroupsuffixO=_o1.fq.gz \\\n    outputperreadgroupsuffixO2=_o2.fq.gz \\\n    outputperreadgroupsuffixS=_s.fq.gz\n\'\'\'\n\npAlignment = Proc(desc = \'Align reads to reference genome.\')\npAlignment.input = \'fq1:file, fq2:file\'\n#                             name_1.fq.gz => name.bam\npAlignment.output = \'bam:file:{{i.fq1 | stem | stem | [:-2]}}.bam\'\npAlignment.script = \'\'\'\nbwa mem -t 8 -T 0 -R <read_group> <reference> {{i.fq1}} {{i.fq2}} | \\\n    samtools view -Shb -o {{o.bam}} -\n\'\'\'\n\npBamSort = Proc(desc = \'Sort bam files.\')\npBamSort.input = \'inbam:file\'\npBamSort.output = \'outbam:file:{{i.inbam | basename}}\'\npBamSort.script = \'\'\'\njava -jar picard.jar SortSam CREATE_INDEX=true INPUT={{i.inbam}} \\\n    OUTPUT={{o.outbam}} SORT_ORDER=coordinate VALIDATION_STRINGENCY=STRICT\n\'\'\'\n\npBamMerge = Proc(desc = \'Merge bam files.\')\npBamMerge.input = \'inbam:file\'\npBamMerge.output = \'outbam:file:{{i.inbam | basename}}\'\npBamMerge.script = \'\'\'\njava -jar picard.jar MergeSamFiles ASSUME_SORTED=false CREATE_INDEX=true \\\n    INPUT={{i.inbam}} MERGE_SEQUENCE_DICTIONARIES=false OUTPUT={{o.outbam}} \\\n    SORT_ORDER=coordinate USE_THREADING=true VALIDATION_STRINGENCY=STRICT\n\'\'\'\n\npMarkDups = Proc(desc = \'Mark duplicates.\')\npMarkDups.input = \'inbam:file\'\npMarkDups.output = \'outbam:file:{{i.inbam | basename}}\'\npMarkDups.script = \'\'\'\njava -jar picard.jar MarkDuplicates CREATE_INDEX=true INPUT={{i.inbam}} \\\n    OUTPUT={{o.outbam}} VALIDATION_STRINGENCY=STRICT\n\'\'\'\n```\n\nEach process is indenpendent so that you may also reuse the processes in other pipelines.\n\n## Pipeline flowchart\n```python\n# When try to run your pipline, instead of:\n#   PyPPL().start(pBamToFastq).run()\n# do:\nPyPPL().start(pBamToFastq).flowchart().run()\n```\nThen an SVG file endswith `.pyppl.svg` will be generated under current directory.\nNote that this function requires [Graphviz][13] and [graphviz for python][12].\n\nSee plugin [details][22].\n\n![flowchart][17]\n\n## Pipeline report\nSee plugin [details][23]\n\n````python\npPyClone.report = """\n## {{title}}\n\nPyClone[1] is a tool using Probabilistic model for inferring clonal population structure from deep NGS sequencing.\n\n![Similarity matrix]({{path.join(job.o.outdir, "plots/loci/similarity_matrix.svg")}})\n\n```table\ncaption: Clusters\nfile: "{{path.join(job.o.outdir, "tables/cluster.tsv")}}"\nrows: 10\n```\n\n[1]: Roth, Andrew, et al. "PyClone: statistical inference of clonal population structure in cancer." Nature methods 11.4 (2014): 396.\n"""\n\n# or use a template file\n\npPyClone.report = "file:/path/to/template.md"\n````\n\n```python\nPyPPL().start(pPyClone).run().report(\'/path/to/report\', title = \'Clonality analysis using PyClone\')\n```\n\n![report][24]\n\n## Full documentation\n[ReadTheDocs][1]\n\n\n[1]: https://pyppl.readthedocs.io/en/latest/\n[2]: https://pyppl.readthedocs.io/en/latest/api/\n[3]: https://github.com/pwwang/pyppl/\n[4]: https://img.shields.io/codacy/grade/30db7165ed6c4d7aa3261cd1f5530813.svg?style=flat-square\n[5]: https://img.shields.io/travis/pwwang/PyPPL.svg?style=flat-square\n[6]: https://img.shields.io/codacy/coverage/30db7165ed6c4d7aa3261cd1f5530813.svg?style=flat-square\n[7]: https://travis-ci.org/pwwang/PyPPL\n[8]: https://app.codacy.com/project/pwwang/PyPPL/dashboard\n[9]: https://pyppl.readthedocs.io/en/latest/CHANGELOG/\n[10]: https://img.shields.io/pypi/v/pyppl.svg?style=flat-square\n[11]: https://img.shields.io/github/tag/pwwang/PyPPL.svg?style=flat-square\n[12]: https://github.com/xflr6/graphviz\n[13]: https://www.graphviz.org/\n[14]: https://img.shields.io/pypi/pyversions/PyPPL.svg?style=flat-square\n[15]: https://docs.gdc.cancer.gov/Data/Bioinformatics_Pipelines/images/dna-alignment-pipeline_0.png\n[16]: https://docs.gdc.cancer.gov/Data/Bioinformatics_Pipelines/DNA_Seq_Variant_Calling_Pipeline/\n[17]: https://raw.githubusercontent.com/pwwang/PyPPL/development/examples/demo/demo.pyppl.svg?sanitize=true\n[18]: https://pypi.org/project/PyPPL/\n[19]: https://img.shields.io/readthedocs/pyppl.svg?style=flat-square\n[20]: https://asciinema.org/a/Uiz6Wdo1buGCGPFd89bWiZzwn.svg?sanitize=true\n[21]: https://asciinema.org/a/Uiz6Wdo1buGCGPFd89bWiZzwn\n[22]: https://github.com/pwwang/pyppl_flowchart\n[23]: https://github.com/pwwang/pyppl_report\n[24]: https://pyppl_report.readthedocs.io/en/latest/snapshot.png\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/PyPPL',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Exceptions for PyPPL"""


class PyPPLInvalidConfigurationKey(KeyError):
    """@API\nWhen invalid configuration key passed"""


class PyPPLFindNoProcesses(Exception):
    """@API\nWhen failed to find any processes with given pattern"""


class PyPPLResumeError(Exception):
    """@API\nTry to resume when no start process has been specified"""


class PyPPLNameError(Exception):
    """@API\nPipeline name duplicated after transformed
    by utils.name2filename"""


class PyPPLWrongPositionMethod(Exception):
    """@API\nWrong position for plugin-added method"""


class PyPPLMethodExists(Exception):
    """@API\nMethod has already been registered"""


class ProcessAttributeError(Exception):
    """@API\nProcess AttributeError"""


class ProcessInputError(Exception):
    """@API\nProcess Input error"""


class ProcessOutputError(Exception):
    """@API\nProcess Output error"""


class ProcessScriptError(Exception):
    """@API\nProcess script building error"""


class ProcessAlreadyRegistered(Exception):
    """@API\nProcess already registered with the same id and tag"""
    def __init__(self, message='', proc1=None, proc2=None):
        """@API
		Construct for ProcessAlreadyRegistered
		@params:
			message (str): The message, make the class to be compatible with Exception
			proc1 (Proc): the first Proc
			proc2 (Proc): the second Proc
		"""
        if not message and not proc1 and not proc2:  # pragma: no cover
            message = 'There are two processes with the same id and tag.'
        elif not message and proc1 and proc2:
            message = '\n'.join([
                "There are two processes with id({}) and tag({})".format(
                    proc1.id, proc1.tag), "", ">>> One is defined here:",
                ''.join(defs for defs in proc1._defs),
                ">>> The other is defined here:",
                ''.join(defs for defs in proc2._defs)
            ])
        super().__init__(message)


class JobInputParseError(Exception):
    """@API\nFailed to parse job input"""


class JobOutputParseError(Exception):
    """@API\nFailed to parse job output"""


class JobBuildingError(Exception):
    """@API\nFailed to build the job"""


class JobFailError(Exception):
    """@API\nJob results validation failed"""


class PluginNoSuchPlugin(Exception):
    """@API\nWhen try to find a plugin not existing"""


class RunnerNoSuchRunner(Exception):
    """@API\nWhen no such runner is found"""


class RunnerMorethanOneRunnerEnabled(Exception):
    """@API\nWhen more than one runners are enabled"""
    def __init__(self, message=''):
        message += ', you may have to call runner.use_runner() first.'
        super().__init__(message)


class RunnerTypeError(Exception):
    """@API\nWrong type of runner"""



"""List all available plugins, including runners,
even they are temporarily disabled."""
from ..plugin import hookimpl, pluginmgr
from ..runner import runnermgr, _runner_name
from ..logger import logger


@hookimpl
def cli_addcmd(commands):
    """Add plugins command"""
    commands.plugins._hbald = False
    commands.plugins = __doc__


@hookimpl
def cli_execcmd(command, opts):  # pylint: disable=unused-argument
    """Run the command"""
    if command == 'plugins':
        logger.initialize({'level': 'notset'})
        for plugin in sorted(pluginmgr.get_plugins(), key=pluginmgr.get_name):

            logger.plugin(
                'Plugin %s: v%s', pluginmgr.get_name(plugin), 'Unknown'
                if not hasattr(plugin, '__version__') else plugin.__version__)

        for plugin in sorted(runnermgr.get_plugins(), key=_runner_name):

            logger.plugin(
                'Runner %s: v%s', _runner_name(plugin), 'Unknown'
                if not hasattr(plugin, '__version__') else plugin.__version__)



"""Job for PyPPL"""
from datetime import datetime
import attr
import toml
from attr_property import attr_property_class, attr_property
from diot import Diot
from .runner import runnermgr, current_runner
from .plugin import pluginmgr, PluginConfig
from .utils import fs, filesig
from ._proc import OUT_DIRTYPE, OUT_STDOUTTYPE, OUT_STDERRTYPE, OUT_VARTYPE
from ._job import (job_dir, job_input, job_output, job_signature,
                   job_rc_setter, job_rc_getter, job_data, job_script,
                   job_logger, job_pid_setter, job_pid_getter,
                   job_script_parts, RC_NO_RCFILE)


@attr_property_class
@attr.s(slots=True)
class Job:
    """@API
    Job class

    @arguments:
        index (int): The index of the job
        proc (Proc): The process
    """

    index = attr.ib()
    proc = attr.ib()

    # allow state machine to add attributes
    __dict__ = attr.ib(init=False, eq=False, repr=False)

    # the job directory
    dir = attr_property(init=False,
                        repr=False,
                        getter=job_dir,
                        doc='The directory of the job.')
    # max number of retries
    ntry = attr.ib(default=0, init=False, eq=False, repr=False)
    # the compiled input
    input = attr_property(init=False, repr=False, eq=False, getter=job_input)
    # the compiled output
    output = attr_property(init=False, repr=False, eq=False, getter=job_output)
    # the compiled script
    script = attr_property(init=False, repr=False, eq=False, getter=job_script)
    # parts of the script for runner
    script_parts = attr_property(init=False,
                                 repr=False,
                                 eq=False,
                                 getter=job_script_parts)
    # the signature of the job
    signature = attr_property(init=False,
                              repr=False,
                              eq=False,
                              getter=job_signature)
    # the returncode of the job
    rc = attr_property(init=False,
                       repr=False,
                       eq=False,
                       setter=job_rc_setter,
                       getter=job_rc_getter,
                       doc='The returncode of the job.')
    # the pid or job id in runner system
    pid = attr_property(init=False,
                        repr=False,
                        eq=False,
                        getter=job_pid_getter,
                        setter=job_pid_setter,
                        doc='The pid/job id of the job in the running system.')
    # current runner name
    # make sure use_runner was called before class definitions
    runner = attr_property(init=False,
                           repr=False,
                           eq=False,
                           getter=lambda this, value: current_runner(),
                           doc='The name of current runner.')
    # data to render templates
    data = attr_property(init=False,
                         repr=False,
                         eq=False,
                         getter=job_data,
                         doc='The data used to render templates.')
    # job logger
    logger = attr_property(init=False, repr=False, eq=False, getter=job_logger)
    # configs for plugins
    config = attr.ib(default=attr.Factory(PluginConfig),
                     eq=False,
                     kw_only=True,
                     repr=False)

    def __attrs_post_init__(self):
        pluginmgr.hook.job_init(job=self)

    def add_config(self, name, default=None, converter=None):
        """@API
        Add a config to plugin config, used for plugins
        @params:
            name (str): The name of the config.
                To proptect your config, better use a prefix
            default (Any): The default value for the config
            converter (callable): The converter for the value
        """
        self.config.add(name, default, converter)

    def cache(self):
        """@API
        Truly cache the job (by signature)
        """
        if self.proc.cache:
            self.signature.to_toml(filename=self.dir / 'job.cache')

    def _is_cached_without_cachefile(self):
        # check if outfiles/dirs are newer than any input files/dirs
        for datatype, data in self.output.values():
            if datatype in OUT_VARTYPE:
                continue
            sig = filesig(data,
                          dirsig=self.proc.dirsig and datatype in OUT_DIRTYPE)
            if sig[1] < self.signature.mtime:
                return False
        return True

    def is_cached(self): # pylint: disable=too-many-return-statements
        """Tell if the job is cached
        @returns:
            (bool): True if the job is cached else False
        """
        if not self.proc.cache or not self.is_succeeded():
            self.logger('Not cached as proc.cache is False or job failed.',
                        slevel="CACHE_FAILED",
                        level='debug')
            return False

        if self.proc.cache == 'force':
            self.cache()
            return True

        if not self.dir.joinpath('job.cache').is_file():
            return self._is_cached_without_cachefile()

        sig_now = self.signature
        with open(self.dir.joinpath('job.cache')) as fcache:
            sig_old = Diot(toml.load(fcache))
        # time is easy to check, check it first
        if sig_now.mtime > sig_old.mtime:
            self.logger(
                'Input or script has been modified (mtime: {} -> {})'.format(
                    datetime.fromtimestamp(sig_old.mtime),
                    datetime.fromtimestamp(sig_now.mtime),
                ),
                slevel="CACHE_INPUT_MODIFIED",
                level='debug')
            return False

        # check items
        if list(sig_now.i.keys()) != list(sig_old.i.keys()):
            additional = [key for key in sig_now.i if key not in sig_old.i]
            if additional:
                self.logger('Additional input items found: %s.' % additional,
                            slevel="CACHE_INPUT_MODIFIED",
                            level='debug')
                return False
            missing = [key for key in sig_old.i if key not in sig_now.i]
            if missing:
                self.logger('Missing input items: %s.' % missing,
                            slevel="CACHE_INPUT_MODIFIED",
                            level='debug')
                return False
        elif sig_now.i != sig_old.i:
            diff_key = [
                key for key in sig_now.i if sig_now.i[key] != sig_old.i[key]
            ][0]
            self.logger('Input item %r changed: %s -> %s' %
                        (diff_key, sig_old.i[diff_key], sig_now.i[diff_key]),
                        slevel="CACHE_INPUT_MODIFIED",
                        level='debug')
            return False

        if list(sig_now.o.keys()) != list(sig_old.o.keys()):
            additional = [key for key in sig_now.o if key not in sig_old.o]
            if additional:
                self.logger('Additional output items found: %s.' % additional,
                            slevel="CACHE_OUTPUT_MODIFIED",
                            level='debug')
                return False
            missing = [key for key in sig_old.o if key not in sig_now.o]
            if missing:
                self.logger('Missing output items: %s.' % missing,
                            slevel="CACHE_OUTPUT_MODIFIED",
                            level='debug')
                return False
        elif sig_now.o != sig_old.o:
            diff_key = [
                key for key in sig_now.o if sig_now.o[key] != sig_old.o[key]
            ][0]
            self.logger('Output item %r changed: %s -> %s' %
                        (diff_key, sig_old.o[diff_key], sig_now.o[diff_key]),
                        slevel="CACHE_OUTPUT_MODIFIED",
                        level='debug')
            return False
        return True

    def build(self):
        """@API
        Initiate a job, make directory and prepare input, output and script.
        """
        self.logger('Builing the job ...', level='debug')
        try:
            # trigger input/output/script building
            self.input  # pylint: disable=pointless-statement
            self.output  # pylint: disable=pointless-statement
            self.script  # pylint: disable=pointless-statement

            pluginmgr.hook.job_prebuild(job=self)
            # check cache
            outfile = self.dir / 'job.stdout'
            errfile = self.dir / 'job.stderr'
            if self.is_cached():
                # we should get stdout/err file back, since job is cached,
                # there is no way to generate new stdout/err,
                # in case they are used somewhere later.
                # we just link them back
                if not fs.exists(outfile):
                    outfile.write_text('')
                if not fs.exists(errfile):
                    errfile.write_text('')
                pluginmgr.hook.job_build(job=self, status='cached')
                return 'cached'

            outfile_bak = self.dir / 'job.stdout.bak'
            errfile_bak = self.dir / 'job.stderr.bak'
            # preserve the outfile and errfile of previous run
            # issue #30
            if fs.exists(outfile):
                fs.move(outfile, outfile_bak)
                outfile.write_text('')
            if fs.exists(errfile):
                fs.move(errfile, errfile_bak)
                errfile.write_text('')

            pluginmgr.hook.job_build(job=self, status='succeeded')
            return True
        except BaseException as ex:
            self.logger('Failed to build job: %s: %s' %
                        (type(ex).__name__, ex),
                        level='debug')
            from traceback import format_exc
            with (self.dir / 'job.stderr').open('w') as ferr:
                ferr.write(str(format_exc()))
            pluginmgr.hook.job_build(job=self, status='failed')
            return False

    def is_succeeded(self):
        """See if the job is successfully done
        Allow plugins to override the status
        By default, we just do a simple check to see if
        the return code is 0
        """
        ret = pluginmgr.hook.job_succeeded(job=self)
        if ret and any(state is False for state in ret):
            return False
        return self.rc == 0

    def done(self, cached=False, status=True):
        """@API
        Do some cleanup when job finished
        @params:
            cached (bool): Whether this is running for a cached job.
        """
        self.logger('Finishing up the job ...', level='debug')

        if cached:
            pluginmgr.hook.job_done(job=self, status='cached')
            return

        if status:
            self.cache()

        pluginmgr.hook.job_done(job=self,
                                status='succeeded' if status else 'failed')

    def reset(self):  # pylint: disable=too-many-branches
        """@API
        Clear the intermediate files and output files"""
        retry = self.ntry
        retrydir = self.dir / ('retry.' + str(retry))
        #cleanup retrydir
        if retry:
            # will be removed by fs.makedirs
            #fs.remove(retrydir)
            fs.makedirs(retrydir)
        else:
            for retrydir in self.dir.glob('retry.*'):
                fs.remove(retrydir)

        for jobfile in ('job.rc', 'job.stdout', 'job.stderr', 'job.pid',
                        'job.cache'):
            if retry and fs.exists(self.dir / jobfile):
                fs.move(self.dir / jobfile, retrydir / jobfile)
            else:
                fs.remove(self.dir / jobfile)
        # try to keep the cache dir, which, in case,
        # if some program can resume from
        if not fs.exists(self.dir / 'output/.jobcache'):
            if retry:
                fs.move(self.dir / 'output', retrydir / 'output')
            else:
                fs.remove(self.dir / 'output')
        elif retry:
            retryoutdir = retrydir / 'output'
            fs.makedirs(retryoutdir)
            # move everything to retrydir but the cachedir
            for outfile in (self.dir / 'output').glob('*'):
                if outfile.name == '.jobcache':
                    continue
                fs.move(outfile, retryoutdir / outfile.name)
        else:
            for outfile in (self.dir / 'output').glob('*'):
                if outfile.name == '.jobcache':
                    continue
                fs.remove(outfile)

        (self.dir / 'job.stdout').write_text('')
        (self.dir / 'job.stderr').write_text('')

        try:
            fs.makedirs(self.dir / 'output', overwrite=False)
        except OSError:
            pass

        for outtype, outdata in self.output.values():
            if outtype in OUT_DIRTYPE:
                # it has been moved to retry dir or removed
                fs.makedirs(outdata)
            if outtype in OUT_STDOUTTYPE:
                fs.link(self.dir / 'job.stdout', outdata)
            if outtype in OUT_STDERRTYPE:
                fs.link(self.dir / 'job.stderr', outdata)

    def submit(self):
        """@API
        Submit the job
        @returns:
            (bool): `True` if succeeds else `False`"""
        self.logger('Submitting the job ...', level='debug')
        # If I am retrying, submit the job anyway.
        if self.ntry == 0 and runnermgr.hook.isrunning(job=self):
            self.logger('is already running at %s, skip submission.' %
                        self.pid,
                        level='SBMTING')
            pluginmgr.hook.job_submit(job=self, status='running')
            return True
        self.reset()
        rscmd = runnermgr.hook.submit(job=self)
        if rscmd.rc == 0:
            pluginmgr.hook.job_submit(job=self, status='succeeded')
            return True
        self.logger(
            'Submission failed (rc = {rscmd.rc}, cmd = {rscmd.cmd})\n'
            '{rscmd.stderr}'.format(rscmd=rscmd),
            slevel='SUBMISSION_FAIL',
            level='error'
        )
        pluginmgr.hook.job_submit(job=self, status='failed')
        return False

    def poll(self):
        """@API
        Check the status of a running job
        @returns:
            (bool|str): `True/False` if rcfile generared and
                whether job succeeds, otherwise returns `running`.
        """
        # force to read rc from rcfile
        del self.rc

        if (not self.dir.joinpath('job.stderr').is_file() or
                not self.dir.joinpath('job.stdout').is_file()):
            self.logger('Polling the job ... stderr/out file not generared.',
                        level='debug')
            pluginmgr.hook.job_poll(job=self, status='running')
            return 'running'

        if self.rc != RC_NO_RCFILE:
            self.logger('Polling the job ... done.', level='debug')
            pluginmgr.hook.job_poll(job=self, status='done')
            return self.is_succeeded()
        # running
        self.logger('Polling the job ... rc file not generated.',
                    level='debug')
        pluginmgr.hook.job_poll(job=self, status='running')
        return 'running'

    def retry(self):
        """@API
        If the job is available to retry
        @returns:
            (bool|str): `ignore` if `errhow` is `ignore`, otherwise
                returns whether we could submit the job to retry.
        """
        if self.proc.errhow == 'ignore':
            return 'ignored'
        if self.proc.errhow != 'retry':
            return False

        self.ntry += 1
        if self.ntry > self.proc.errntry:
            return False

        self.logger('Retrying {} out of {} time(s) ...'.format(
            str(self.ntry).rjust(len(str(self.proc.errntry)), ' '),
            self.proc.errntry),
                    level='rtrying')
        return True

    def kill(self):
        """@API
        Kill the job
        @returns:
            (bool): `True` if succeeds else `False`
        """
        self.logger('Killing the job ...', level='debug')
        try:
            status = runnermgr.hook.kill(job=self)
        except BaseException:
            self.pid = ''
            pluginmgr.hook.job_poll(job=self, status='failed')
            return False

        if status:
            self.pid = ''
        pluginmgr.hook.job_kill(job=self,
                                status='succeeded' if status else 'failed')
        return status
